@@ -407,6 +407,11 @@ NOEXPORT void ssl_start(CLI *c) {
         SSL_set_accept_state(c->ssl);
     }
 
+    if(c->opt->option.require_cert)
+        s_log(LOG_INFO, "Peer certificate required");
+    else
+        s_log(LOG_INFO, "Peer certificate not required");
+
     unsafe_openssl=OpenSSL_version_num()<0x0090810fL ||
         (OpenSSL_version_num()>=0x10000000L &&
         OpenSSL_version_num()<0x1000002fL);
@@ -483,7 +488,7 @@ NOEXPORT void ssl_start(CLI *c) {
         peer_cert=SSL_get_peer_certificate(c->ssl);
         if(peer_cert) /* c->redirect was set by the callback */
             X509_free(peer_cert);
-        else if(c->opt->redirect_addr.names)
+        else if(c->opt->redirect_addr.names) /* no peer certificate verified */
             c->redirect=REDIRECT_ON;
         SSL_SESSION_set_ex_data(SSL_get_session(c->ssl),
             index_redirect, (void *)c->redirect);
