@@ -8,13 +8,10 @@ fi
 
 if test -n "$2"; then
     OPENSSL="$2/bin/openssl"
-    export LD_LIBRARY_PATH="$2/lib:$2/lib64"
 else
     OPENSSL=openssl
 fi
 
-$OPENSSL version
-
 if test -n "$3"; then
     RAND="$3"
 else
